#!/usr/bin/perl -w

use strict;
use Getopt::Long;

use POSIX qw(locale_h);
use locale;
use IO::Handle;
STDOUT->autoflush(1);


sub permut
{
    my ( @array ) = @_;
    my $size = scalar(@array);
    for( my $i=0 ; $i<$size ; $i++ )
    {
	my $r = rand( $size - $i ) + $i;
	( $array[$i], $array[$r] ) = ( $array[$r], $array[$i] );
    }
    return @array;
}


sub letter_tree_add
{
    my ( $tree, $letters, $words ) = @_;
    if( $letters eq '' )
    {
        $tree->{'.'} = $words;
    }
    else
    {
        my $l = substr( $letters, 0, 1 );
	if( ! defined $tree->{$l} )
	{
	    my %h;
	    $tree->{$l} = \%h;
	}
    	letter_tree_add( $tree->{$l}, substr( $letters, 1 ), $words );
    }
}

sub letter_tree_included_rec
{
    my ( $tree, $letters, $n ) = @_;
    my @words;
    if( defined $tree->{'.'} )
    {
	@words = @{$tree->{'.'}};
    }
    for( my $i=0 ; $i<$n ; $i++ )
    {
	my $l = substr( $letters, $i, 1 );
	if( defined $tree->{$l} )
	{
	    @words = ( @words, letter_tree_included( $tree->{$l}, substr($letters,$i+1), $n-1 ) );
	}
    }
    return @words;
}

sub letter_tree_included
{
    my ( $tree, $letters ) = @_;    
    return letter_tree_included_rec( $tree, $letters, length($letters) );
}


sub play
{
    my ( $letter, @list ) = @_;
    my @list_;
    my $count = scalar(@list);
    my $number = $count; 
    
    foreach my $w (@list)
    {
	@list_ = ( @list_, "-" x length($w) );
    }
    
    while( 1 )
    {
	print "letters: $letter\n";
	print "words($count/$number): ".join(".",@list_)."\n";
	my $old_count = $count;
	my $text = <STDIN>;
	if( $text )
	{
	    $_ = $text;
	    if( /^\:/ )
	    {
		if( /^\:quit$/ )
		{
		    return 0;
		}
		elsif( /^\:new$/ )
		{
		    return 1;
		}
		elsif( /^\:help$/ && $count>0 )
		{
		    my $n;
		    do { $n = rand($number); } while( !( $list_[$n] =~ /.*-.*/ ) );
		    my $m;
		    do { $m = rand(length($list_[$n])); } while( substr($list_[$n],$m,1) ne '-' );
		    $list_[$n] = substr($list_[$n],0,$m) . substr($list[$n],$m,1) . substr($list_[$n],$m+1);
		    if( !( $list_[$n] =~ /.*-.*/ ) )
		    {
			$count--;
		    }
		}
		elsif( /^\:resolve$/ && $count>0 )
		{
		    for( my $i=0 ; $i<$number ; $i++ )
		    {
		        $list_[$i] = $list[$i];
		    }
		    $count = 0;
		}
	    }
	    else
	    {
		chomp( $text );
		for( my $i=0 ; $i<$number ; $i++ )
		{
		    $_ = $list[$i];
		    tr//aaaaaeeeeiiiiooooouuuuyycn/;
		    my $l = $_;
		    $_ = $list_[$i];
		    tr//aaaaaeeeeiiiiooooouuuuyycn/;
		    my $l_ = $_;
		    if( $l eq $text && $l_ ne $text)
		    {
		        $list_[$i] = $list[$i];
			$count--;
		    }
		}
	    }
	}
	print "result: ".(($count==0)? "end" : ($old_count==$count)? "fail" : "ok")."\n";
    }
}

sub load_dict
{
    my ( $file_ana, $min_size, $max_size ) = @_;
    open( ANA, "$file_ana" ) || die "Can't open \"$file_ana\"\n";
    
    my ( %ret, @ret_maxsize ) = ( (), () );
    while( <ANA> )
    {
	chomp;
	my @e = split( /:/, $_ );
	if( length($e[0]) >= $min_size )
	{
	    if( length($e[0]) < $max_size )
	    {
		my @words = split( /,/, $e[1] );
		letter_tree_add( \%ret, $e[0], \@words );
	    }
	    elsif( length($e[0]) == $max_size )
	    {
		push( @ret_maxsize, \@e );
	    }
	}
    }
    close( ANA );

    return ( \%ret, \@ret_maxsize );
}




my ( $help, $list_dictionary );
my ( $dictionary, $min_size, $max_size, $min_number ) = ( 'en', 4, 7, 5 );
GetOptions( 'help|?'=>\$help, 'list-dictionary'=>\$list_dictionary, 'dictionary=s'=>\$dictionary, 'min-size=i'=>\$min_size, 'max-size=i'=>\$max_size, 'min-number=i'=>\$min_number );
if( defined $help )
{
    print STDERR "Usage: r-anagram.perl [-h|--help] [-l|--list-dictionary] [-d|--dictionary lang ($dictionary)] [--min-size int ($min_size)] [--max-size int ($max_size)] [--min-number int ($min_number)]\n";
    print "init: exit\n";
}
elsif( defined $list_dictionary )
{
    opendir( DIR, "../dictionary" );
    while( $_=readdir(DIR) )
    {
	if( /(.*)\.dict\.anagram/ )
	{
	    print STDERR "$1\n";
	}
    }
    closedir( DIR );
    print "init: exit\n";
}
else
{
    my ( $dict_, $dict_maxsize_ ) = load_dict( '../dictionary/'.$dictionary.'.dict.anagram', $min_size, $max_size );
    if( scalar($dict_maxsize_) > 0 )
    {
	print "init: ready\n";
	$_ = $dictionary; my ($locale) = /(..)/;
	setlocale( LC_CTYPE, $locale );
    }
    else
    {
	print "init: Can't init \"$dictionary\" with min-size=$min_size and max-size=$max_size\n";
	exit;
    }

    my @dict_maxsize = permut( @$dict_maxsize_ );

    my @words;
    my $letter;
    my $n = 0;
    my $have_anagram = 0;
    do
    {
	do
	{
	    my $words;
	    ( $letter, $words ) = @{$dict_maxsize[$n]};
	    my @list = split( /,/, $words );
	    @list = ( @list, letter_tree_included( $dict_, $letter ) );
	    my @words_ = sort( {if(length($a)!=length($b)){length($a)<=>length($b)}else{$a cmp $b}} @list );
	    @words = ( "" ); foreach my $w (reverse(@words_)) { if( $words[0] ne $w ) { @words = ( $w, @words ); } }; pop( @words );
	    if( $n < scalar(@dict_maxsize)-1 )
	    {
		$n++;
	    }
	    else
	    {
		if( $have_anagram == 0 )
		{
		    print STDERR "Fail creating anagram with \"$dictionary\", min-size=$min_size and max-size=$max_size\n";
		    exit;
		}
		$n = 0;
	    }
	}
	while( scalar(@words) < $min_number || scalar(@words) > 15 );
	$have_anagram = 1;
    }
    while( play( $letter, @words ) );
}
