#include "letters_word.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "sprite_words.h"
#include "locate_data.h"

letters_word::letters_word( char *letters, char **words, set<Sprite*> *sprite_set ) :
sprite_set( sprite_set )
{
    this->letters = strdup( letters );
    size = strlen(letters);
    word = (char*) malloc( sizeof(char) * (size+1) );
    word[0] = 0;
    sprite_letters = (Sprite_letter**) malloc( sizeof(Sprite_letter*) * size );
    sprite_word    = (Sprite_letter**) malloc( sizeof(Sprite_letter*) * size );

    font_letter = TTF_OpenFont( locate_data("font/luxisr.ttf"), 52 );
    assert( font_letter != NULL );
    font_words = TTF_OpenFont( locate_data("font/karine.ttf"), 36 );
    assert( font_words != NULL );

    for( int i=0 ; i<size ; i++ )
    {
//        sprite_letters[i] = new Sprite_letter( (int)(((float)rand())/RAND_MAX*(300)), 0, letters[i], font_letter, "images/[].png" );
//	sprite_letters[i]->slideTo( i*64+200, 256 );
        sprite_letters[i] = new Sprite_letter( i*64+200, 256, letters[i], font_letter, "images/[].png" );
        sprite_set->insert( sprite_letters[i] );
        sprite_word[i] = NULL;
    }
    sprite_words = new Sprite_words( 300, 600-111, "images/paper.png", font_words, words );
    sprite_set->insert( sprite_words );
}

letters_word::~letters_word()
{
    free( letters );
    free( word );
    for( int i=0 ; i<size ; i++ )
    {
        if( sprite_letters[i] != NULL )
	{
//	    sprite_letters[i]->slideTo( 800-(int)(((float)rand())/RAND_MAX*(300)), 0 );
	    sprite_set->erase( sprite_letters[i] );
            delete sprite_letters[i];
	}
        if( sprite_word   [i] != NULL )
	{
//	    sprite_letters[i]->slideTo( 800-(int)(((float)rand())/RAND_MAX*(300)), 0 );
	    sprite_set->erase( sprite_word[i] );
            delete sprite_word[i];
	}
    }
    free( sprite_letters );
    free( sprite_word );
    sprite_set->erase( sprite_words );
    delete sprite_words;
    TTF_CloseFont( font_letter );
    TTF_CloseFont( font_words );
}


int letters_word::letters_click( int x, int y )
{
    for( int i=0 ; i<size ; i++ )
        if( sprite_letters[i]!=NULL && sprite_letters[i]->click( x, y ) )
	    return i;
    return -1;
}		

int letters_word::word_click( int x, int y )
{
    for( int i=0 ; sprite_word[i]!=NULL && i<size ; i++ )
        if( sprite_word[i]->click( x, y ) )
	    return i;
    return -1;
}


bool letters_word::move_letter_to_word( char letter, int n  )
{
    bool valid = false;
    int i;
    if( n >= 0 && n<size )
    {
	i = n;
	valid = true;
    }
    else
	for( i=0 ; !valid && i<size ; i++ )
	{
	    if( letters[i] == letter )
	    {
		valid = true;
		break;
	    }
	}

    if( valid )
    {
	int l = strlen( word );
	word[l] = letters[i];
	word[l+1] = 0;
	letters[i] = ' ';
	sprite_word[l] = sprite_letters[i];
	sprite_letters[i] = NULL;
	sprite_word[l]->slideTo( l*64+200, 128 );
	sprite_word[l]->run();
	return true;
    }
    return false;
}

bool letters_word::move_letter_from_word( int n, bool slide )
{
    int l = strlen( word );
    if( l > 0 )
    {
	if( n >= 0 )
	{
	    if( n < l )
		l = n;
	    else
		return false;
	}
	else
	{
	    l--;
	}
	int i;
	for( i=0 ; letters[i] != ' ' ; i++ );
	letters[i] = word[l];
	sprite_letters[i] = sprite_word[l];
	if( slide )
	    sprite_letters[i]->slideTo( i*64+200, 256 );
	else
	    sprite_letters[i]->moveTo( i*64+200, 256 );
	//sprite_letters[i]->run();
	for( int j=l ; word[j]!=0 ; j++ )
	{
	    word[j] = word[j+1];
	    
	    if( j+1 < size )
	    {
		sprite_word[j] = sprite_word[j+1];
		if( sprite_word[j] != NULL )
		    if( slide )
			sprite_word[j]->slideTo( j*64+200, 128 );
		    else
			sprite_word[j]->moveTo( j*64+200, 128 );		    
	    }
	    else
	    {
		sprite_word[j] = NULL;
	    }
	}
	return true;
    }
    else
    {
	return false;
    }
}

void letters_word::shuffle_letters()
{
    clear_word();

    for( int i=0 ; i<size ; i++ )
    {
	int j = (int)(((float)rand())/RAND_MAX*(size-i))+i;
	char tmp = letters[i];
	letters[i] = letters[j];
	letters[j] = tmp;
	Sprite_letter *tmps = sprite_letters[i];
	sprite_letters[i] = sprite_letters[j];
	sprite_letters[j] = tmps;
    }

    for( int i=0 ; i<size ; i++ )
    {
	sprite_letters[i]->slideTo( i*64+200, 256 );
	//sprite_letters[i]->run();
    }
}

void letters_word::clear_word()
{
    while( move_letter_from_word( -1, true ) );
}

void letters_word::clean_word()
{
    while( move_letter_from_word( -1, false ) );
}

void letters_word::set_words( char **words )
{
    sprite_words->setWords( words );
}
