#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>

#include <set.h>

#include "gettext.h"
#include "locate_data.h"

#include "SDL.h"
#include "SDL_image.h"
#include "SDL_ttf.h"

#include "SDL_framerate.h"

#include "engine.h"
#include "engine_io.h"
#include "sprite.h"
#include "sprite_text.h"
#include "letters_word.h"

SDL_Surface *screen;

int FPS=60;
FPSmanager FPS_manager;

typedef enum { action_null, action_valid, action_erase, action_shuffle, action_quit, action_type, action_new, action_help, action_resolve, action_untype } action;

struct SDL_Rect_cmp { bool operator()( const SDL_Rect &r1, const SDL_Rect &r2 ) const { return &r1<&r2; } };


void refresh_screen( SDL_Surface *screen, SDL_Surface *background, set<Sprite*> &sprite_set )
{
    int x, y;
    SDL_GetMouseState( &x, &y );
    set<SDL_Rect,SDL_Rect_cmp> out_of_date;
    for( set<Sprite*>::iterator i=sprite_set.begin() ; i!=sprite_set.end() ; i++ )
        (*i)->hover( x, y );
    for( set<Sprite*>::iterator i=sprite_set.begin() ; i!=sprite_set.end() ; i++ )
    {
        SDL_Rect splash;
        if( (*i)->clear( &splash ) )
	{
    	    out_of_date.insert( splash );
	    SDL_BlitSurface( background, &splash, screen, &splash );
	}
    }

    for( set<Sprite*>::iterator i=sprite_set.begin() ; i!=sprite_set.end() ; i++ )
    {
	SDL_Rect splash;
	if( (*i)->redraw( &splash ) )
	    out_of_date.insert( splash );
    }
    
    if( ! out_of_date.empty() )
    {
    	for( set<SDL_Rect,SDL_Rect_cmp>::iterator i=out_of_date.begin() ; i!=out_of_date.end() ; i++ )
	{
	    SDL_SetClipRect( screen, &(*i) );
	    for( set<Sprite*>::iterator j=sprite_set.begin() ; j!=sprite_set.end() ; j++ )
		(*j)->blit();
	}
	SDL_SetClipRect( screen, NULL );

        for( set<SDL_Rect,SDL_Rect_cmp>::iterator i=out_of_date.begin() ; i!=out_of_date.end() ; i++ )
	{
	    SDL_UpdateRect( screen, (*i).x<0?0:(*i).x, (*i).y<0?0:(*i).y, (*i).w, (*i).h );
	}
    }
}


void get_data_from_engine( char **letters, char ***words )
{
    if( letters != NULL )
    {
	if( *letters != NULL )
	    free( *letters );
    }
    if( *words != NULL )
    {
	for( int i=0 ; (*words)[i]!=NULL ; i++ )
	    free( (*words)[i] );
	free( *words );
    }

    char *lets = get_letters_from_engine();
    if( letters != NULL )
    {
	*letters = lets;
	assert( *letters != NULL );
	//printf("L:%s\n", *letters );
    }
    else
    {
	if( lets != NULL ) free( lets );
    }

    *words = get_words_from_engine();
    assert( *words != NULL );
    for( int i=0 ; (*words)[i]!=NULL ; i++ )
    {
	//printf( "W[%i]:%s\n", i, (*words)[i] );
    }
}

int main( int argn, char **args )
{
    setlocale( LC_CTYPE, "" );
    setlocale( LC_MESSAGES, "" );

    char *po_dir = LOCALEDIR;
    if( locate_data( "../po", false ) != NULL )
	po_dir = "../po";
    bindtextdomain( PACKAGE, po_dir );
    bind_textdomain_codeset( PACKAGE, "UTF-8" );
    textdomain( PACKAGE );

    engine_start( args );
    char *result = get_answer_result_from_engine();
    assert( result != NULL );
    if( strcmp( result, "exit" ) == 0 )
    {
	free( result );
	return 0;
    }
    else if( strcmp( result, "ready" ) != 0 )
    {
	printf( "%s:engine error:%s\n", args[0], result );
	free( result );
	return 1;
    }
    free( result );

    assert( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE ) == 0 );
    SDL_Surface *icon = IMG_Load( locate_data("r-anagram.png") );
    assert( icon != NULL );
    SDL_WM_SetIcon( icon, NULL );
    SDL_FreeSurface( icon );
    SDL_WM_SetCaption( _("R-anagram"), _("R-anagram") );

    TTF_Init();
    SDL_EnableUNICODE( 1 );
    SDL_initFramerate( &FPS_manager );

    SDL_setFramerate( &FPS_manager, FPS );

    screen = SDL_SetVideoMode( 800, 600, 16, SDL_SWSURFACE  );
    assert( screen != NULL );

    #ifdef AUDIO_SDL
    bool have_audio = ( SDL_InitSubSystem( SDL_INIT_AUDIO ) == 0 );
    if( have_audio )
	have_audio = ( Mix_OpenAudio( MIX_DEFAULT_FREQUENCY, MIX_DEFAULT_FORMAT, 1, 4096 ) == 0 );
    Mix_Chunk *sound_ok = NULL, *sound_fail = NULL;
    Mix_Chunk *sound_type_ok = NULL, *sound_type_fail = NULL;
    Mix_Chunk *sound_win = NULL;
    if( have_audio )
    {
	sound_ok   = Mix_LoadWAV( locate_data("audio/ok.ogg") );
	sound_fail = Mix_LoadWAV( locate_data("audio/fail.ogg") );
	sound_type_ok   = Mix_LoadWAV( locate_data("audio/type_ok.ogg") );
	sound_type_fail = Mix_LoadWAV( locate_data("audio/type_fail.ogg") );
	sound_win = Mix_LoadWAV( locate_data("audio/win.ogg") );
    }
    #else
    bool have_audio = false;
    #endif

    SDL_Surface *background = IMG_Load( locate_data("images/background.png") );
    assert( background != NULL );

    Sprite::screen     = screen;

    TTF_Font *font_menu = TTF_OpenFont( locate_data("font/luxisr.ttf"), 36 );
    assert( font_menu != NULL );

    SDL_BlitSurface( background, NULL, screen, NULL );
    SDL_UpdateRect( screen, 0, 0, 0, 0 );

    set<Sprite*> sprite_set;
    Sprite sprite_valid( 700, 96, "images/valid.png", "images/valid.hover.png", "images/valid.click.png" );
    Sprite sprite_erase( 700, 160, "images/erase.png", "images/erase.hover.png", "images/erase.click.png", have_audio?"audio/erase.ogg":NULL );
    Sprite sprite_shuffle( 700, 256, "images/shuffle.png", "images/shuffle.hover.png", "images/shuffle.click.png", have_audio?"audio/shuffle.ogg":NULL );
    Sprite_text sprite_help( 700, 400, _("help"), font_menu );
    Sprite_text sprite_resolve( 700, 450, _("resolve"), font_menu );
    Sprite_text sprite_new( 700, 500, _("new"), font_menu );
    Sprite_text sprite_quit( 700, 550, _("quit"), font_menu );
    sprite_set.insert( &sprite_valid );
    sprite_set.insert( &sprite_erase );
    sprite_set.insert( &sprite_shuffle );
    sprite_set.insert( &sprite_help );
    sprite_set.insert( &sprite_resolve );
    sprite_set.insert( &sprite_new );
    sprite_set.insert( &sprite_quit );

bool quit = false;
while( !quit )
{
    char *letters = NULL;
    char **words = NULL;
    get_data_from_engine( &letters, &words );

    letters_word lw( letters, words, &sprite_set );

    bool end_play = false;
    while( !quit && !end_play )
    {

	SDL_Event event;
	while( SDL_PollEvent( &event ) )
	{
	    action action = action_null;
	    char type = 0;
	    int type_n = -1;
	    switch( event.type )
	    {
		case SDL_QUIT:
		    action = action_quit;
		    break;
		case SDL_MOUSEBUTTONDOWN:
		    if( sprite_valid.click( event.button.x, event.button.y ) ) action = action_valid;
		    else if( sprite_erase.click( event.button.x, event.button.y ) ) action = action_erase;
		    else if( sprite_shuffle.click( event.button.x, event.button.y ) ) action = action_shuffle;
		    else if( sprite_help.click( event.button.x, event.button.y ) ) action = action_help;
		    else if( sprite_resolve.click( event.button.x, event.button.y ) ) action = action_resolve;
		    else if( sprite_new.click( event.button.x, event.button.y ) ) action = action_new;
		    else if( sprite_quit.click( event.button.x, event.button.y ) ) action = action_quit;
		    else if( (type_n=lw.letters_click( event.button.x, event.button.y )) != -1 ) action = action_type;
		    else if( (type_n=lw.word_click( event.button.x, event.button.y )) != -1 ) action = action_untype;
		    break;
		case SDL_KEYDOWN:
		    switch( event.key.keysym.sym )
		    {
			case SDLK_SPACE:
			    action = action_shuffle;
			    break;
			case SDLK_ESCAPE:
			    action = action_quit;
			    break;
			case SDLK_BACKSPACE:
			    action = action_untype;
			    break;
			case SDLK_RETURN:
			    action = action_valid;
			    break;
			default:
			    if( (event.key.keysym.unicode & 0xFF80) == 0 )
			    {
				char ch = event.key.keysym.unicode & 0x7F;
				if( isalpha( ch ) )
				{
				    action = action_type;
				    type = ch;
				}
			    }
			    break;
		    }
		    break;
		default:
		    break;
	    }
	    switch( action )
	    {
		case action_valid:
		    {
			send_answer_to_engine( lw.get_word() );
			char *result = get_answer_result_from_engine();
			if( strcmp( result, "end" ) == 0 )
			{
			    #ifdef AUDIO_SDL			
			    if( have_audio && sound_win != NULL ) Mix_PlayChannel( -1, sound_win, 0 );
			    #endif			    
			    printf( "END\n" );
			}
			else
			{
			    if( strcmp( result, "ok" ) == 0 )
			    {
				#ifdef AUDIO_SDL
				if( have_audio && sound_ok != NULL ) Mix_PlayChannel( -1, sound_ok, 0 );
				#endif
				printf( "OK\n" );
				while( lw.move_letter_from_word( -1 ) );
			    }
			    else // fail
			    {
				#ifdef AUDIO_SDL
				if( have_audio && sound_fail != NULL ) Mix_PlayChannel( -1, sound_fail, 0 );
				#endif
				printf( "FAIL\n" );
			    }
			}
			if( result != NULL ) free( result );			
			get_data_from_engine( NULL, &words );
			lw.set_words( words );
		    }
		    lw.clear_word();
		    sprite_valid.run();
		    break;
		case action_erase:
		    lw.clear_word();
		    sprite_erase.run();
		    break;
		case action_shuffle:
		    lw.shuffle_letters();
		    sprite_shuffle.run();
		    break;
		case action_help:
		    {
			send_answer_to_engine( ":help" );
			char *result = get_answer_result_from_engine();
			if( result != NULL ) free( result );
			get_data_from_engine( NULL, &words );
			lw.set_words( words );
		    }
		    sprite_help.run();
		    break;
		case action_resolve:
		    {
			send_answer_to_engine( ":resolve" );
			char *result = get_answer_result_from_engine();
			if( result != NULL ) free( result );
			get_data_from_engine( NULL, &words );
			lw.set_words( words );
		    }
		    #ifdef AUDIO_SDL
		    if( have_audio && sound_win != NULL ) Mix_PlayChannel( -1, sound_win, 0 );
		    #endif
		    sprite_resolve.run();
		    break;
		case action_new:
		    send_answer_to_engine( ":new" );
		    lw.clean_word();
		    sprite_new.run();
		    #ifdef AUDIO_SDL
		    if( have_audio && sound_win != NULL ) Mix_PlayChannel( -1, sound_win, 0 );
		    #endif
		    end_play = 1;
		    break;
		case action_quit:
		    sprite_quit.run();
		    quit = 1;
		    break;
		case action_type:
/* not ready		    {
			char table_accent[] = "";
			char table[]        = "aaaaaeeeeiiiiooooouuuuyycn";
			int i;
			for( i=0 ; type!=table_accent[i] && table_accent[i]!=0 ; i++ );
			if( type == table_accent[i] )
			    type = table[i];
		    }*/
		    if( lw.move_letter_to_word( type, type_n ) )
		    {
			#ifdef AUDIO_SDL
			if( have_audio && sound_type_ok != NULL ) Mix_PlayChannel( -1, sound_type_ok, 0 );
			#endif
		    }
		    else
		    {
			#ifdef AUDIO_SDL
			if( have_audio && sound_type_fail != NULL ) Mix_PlayChannel( -1, sound_type_fail, 0 );
			#endif
    		    }
		    break;
		case action_untype:
		    if( lw.move_letter_from_word( type_n ) )
		    {
			sprite_erase.run();
		    }
		    break;
		default:
		    break;
	    }
	}

	refresh_screen( screen, background, sprite_set );
	SDL_framerateDelay( &FPS_manager );
    }

    free( letters );
    for( int i=0 ; words[i]!=NULL ; i++ )
	free( words[i] );
    free( words );
}

    SDL_FreeSurface( background );
    TTF_CloseFont( font_menu );

    #ifdef AUDIO_SDL
    if( sound_ok   != NULL ) Mix_FreeChunk( sound_ok );
    if( sound_fail != NULL ) Mix_FreeChunk( sound_fail );
    if( sound_type_ok   != NULL ) Mix_FreeChunk( sound_type_ok );
    if( sound_type_fail != NULL ) Mix_FreeChunk( sound_type_fail );
    if( sound_win != NULL ) Mix_FreeChunk( sound_win );

    if( have_audio ) Mix_CloseAudio();    
    #endif

    TTF_Quit();
    SDL_Quit();

    return 0;
}
